"""Evaluation window for a Project."""

from __future__ import annotations

import tkinter as tk
from logging import getLogger
from pathlib import Path
from tkinter import messagebox, ttk
from tkinter.scrolledtext import ScrolledText
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt

from scalewiz.components.evaluation_data_view import EvaluationDataView
from scalewiz.components.evaluation_plot_view import EvaluationPlotView
from scalewiz.helpers.export import export
from scalewiz.helpers.score import score
from scalewiz.helpers.set_icon import set_icon
from scalewiz.models.project import Project

if TYPE_CHECKING:

    from scalewiz.models.test_handler import TestHandler


LOGGER = getLogger("scalewiz")


class EvaluationWindow(tk.Toplevel):
    """Frame for analyzing data."""

    def __init__(self, handler: TestHandler) -> None:
        super().__init__()
        self.handler = handler
        self.editor_project = Project()
        if Path(self.handler.project.path.get()).is_file():
            self.editor_project.load_json(self.handler.project.path.get())
        # matplotlib uses these later
        self.log_text: ScrolledText = None
        self.plot_view: EvaluationPlotView = None  # this gets destroyed in plot()
        self.title(f"{self.handler.name} {self.handler.project.name.get()}")
        self.resizable(0, 0)
        set_icon(self)
        self.build()

    def build(self, reload: bool = False) -> None:
        """Destroys all child widgets, then builds the UI."""
        if not self.winfo_exists():
            return

        if reload and Path(self.handler.project.path.get()).is_file():
            # cleanup for the GC
            self.editor_project.remove_traces()
            self.editor_project = Project()
            self.editor_project.load_json(self.handler.project.path.get())

        for child in self.winfo_children():
            if child.winfo_exists():
                self.after(0, child.destroy)

        self.grid_columnconfigure(0, weight=1)
        # we will build a few tabs in this
        self.tab_control = ttk.Notebook(self)
        self.tab_control.grid(row=0, column=0)

        data_view = EvaluationDataView(self.tab_control, self.editor_project)
        self.tab_control.add(data_view, text="   Data   ")

        # plot stuff ----------------------------------------------------------

        # evaluation stuff ----------------------------------------------------
        self.log_frame = ttk.Frame(self.tab_control)
        self.log_frame.grid_columnconfigure(0, weight=1)
        self.log_frame.grid_rowconfigure(0, weight=1)
        self.log_text = ScrolledText(
            self.log_frame, background="white", state="disabled"
        )
        self.log_text.grid(sticky="nsew")
        self.tab_control.add(self.log_frame, text="   Calculations   ")
        self.plot()
        # finished adding to tab control

        button_frame = ttk.Frame(self)
        if self.handler.is_running:
            state = "disabled"
        else:
            state = "normal"
        save_btn = ttk.Button(
            button_frame, text="Save", command=self.save, width=10, state=state
        )
        save_btn.grid(row=0, column=0, padx=5)
        export_btn = ttk.Button(
            button_frame,
            text="Export",
            command=lambda: export(self.editor_project),
            width=10,
        )
        export_btn.grid(row=0, column=1, padx=5)
        button_frame.grid(row=1, column=0, pady=5)
        # update results
        score(self.editor_project, self.log_text)

    def plot(self) -> None:
        """Destroys the old plot frame if it exists, then makes a new one."""
        if isinstance(self.plot_view, EvaluationPlotView):
            plt.close(self.plot_view.fig)
            self.plot_view.destroy()

        self.plot_view = EvaluationPlotView(self.tab_control, self.editor_project)
        self.tab_control.add(self.plot_view, text="   Plot   ")

    def save(self) -> None:
        """Saves to file the project, most recent plot, and calculations log."""
        # update image
        plot_output = (
            f"{self.editor_project.numbers.get().replace(' ', '')} "
            f"{self.editor_project.name.get()} "
            "Scale Block Analysis (Graph).png"
        )
        parent_dir = Path(self.editor_project.path.get()).parent
        plot_output = Path(parent_dir, plot_output).resolve()
        self.plot_view.fig.savefig(plot_output)
        self.editor_project.plot.set(str(plot_output))
        # update log
        log_output = (
            f"{self.editor_project.numbers.get().replace(' ', '')} "
            f"{self.editor_project.name.get()} "
            "Scale Block Analysis (Log).txt"
        ).strip()
        log_output = Path(parent_dir, log_output).resolve()

        with log_output.open("w") as file:
            file.write(self.log_text.get("1.0", "end-1c"))

        # refresh
        self.editor_project.dump_json()
        self.handler.load_project(self.editor_project.path.get())
        self.after(0, self.handler.rebuild_views)

    def export(self) -> None:
        result, file = export(self.editor_project)
        if result == 0:
            messagebox.showinfo("Export complete", f"Exported a report to {file}")
        else:
            messagebox.showwarning(
                "Export failed",
                (
                    f"Failed to export the report to {file}. \n"
                    "Check the log for a more detailed message."
                ),
            )



"""A table view to be displayed in the Evaluation Window."""

from __future__ import annotations

import tkinter as tk
from logging import Logger, getLogger
from tkinter import messagebox, ttk
from tkinter.font import Font
from typing import TYPE_CHECKING

from scalewiz.components.evaluation_plot_view import EvaluationPlotView
from scalewiz.helpers.score import score

if TYPE_CHECKING:
    from typing import List

    from scalewiz.components.evaluation_window import EvaluationWindow
    from scalewiz.models.project import Project
    from scalewiz.models.test import Test

LOGGER: Logger = getLogger("scalewiz")


class EvaluationDataView(ttk.Frame):
    """A widget for selecting devices."""

    def __init__(self, parent: ttk.Frame, project: Project) -> None:
        super().__init__(parent)
        self.eval_window: EvaluationWindow = parent.master
        self.project = project
        self.trials: List[Test] = []
        self.blanks: List[Test] = []
        self.bold_font: Font = Font(family="Arial", weight="bold", size=10)
        self.build()

    def build(self) -> None:
        """Build the UI."""
        for child in self.winfo_children():
            self.after(0, child.destroy)
        self.sort_tests()

        self.apply_col_headers()  # row 0
        # add blanks block
        blanks_lbl = ttk.Label(self, text="Blanks:", font=self.bold_font)
        blanks_lbl.grid(row=1, column=0, sticky="w")
        for i, blank in enumerate(self.blanks):
            self.apply_test_row(blank, i + 2)  # skips rows for headers
        # add trials block
        len_blanks = len(self.blanks)
        trials_lbl = ttk.Label(self, text="Trials:", font=self.bold_font)
        trials_lbl.grid(row=len_blanks + 3, sticky="w")  # skips rows for headers
        for i, trial in enumerate(self.trials):
            self.apply_test_row(trial, i + len_blanks + 4)  # skips rows for headers
        self.update_score()

    def apply_col_headers(self, row: int = 0) -> None:
        """Insert header labels on the passed row."""
        labels = []
        labels.append(
            tk.Label(
                self,
                text="Name",
                font=self.bold_font,
                anchor="w",
            )
        )

        labels.append(
            tk.Label(
                self,
                text="Minutes",
                font=self.bold_font,
                anchor="center",
            )
        )
        labels.append(tk.Label(self, text="Pump", font=self.bold_font, anchor="center"))
        labels.append(
            tk.Label(self, text="Baseline PSI", font=self.bold_font, anchor="center")
        )
        labels.append(
            tk.Label(self, text="Max PSI", font=self.bold_font, anchor="center")
        )
        labels.append(
            tk.Label(self, text="Water Clarity", font=self.bold_font, anchor="center")
        )
        labels.append(tk.Label(self, text="Notes", font=self.bold_font, anchor="w"))
        labels.append(tk.Label(self, text="Score", font=self.bold_font, anchor="w"))
        labels.append(tk.Label(self, text="On Report", font=self.bold_font, anchor="w"))
        # extra for del button row
        labels.append(tk.Label(self, text=" ", font=self.bold_font, anchor="w"))

        for i, lbl in enumerate(labels):
            self.grid_columnconfigure(i, weight=1)
            if i in (0, 1, 7):
                lbl.grid(row=row, column=i, padx=0, sticky="w")
            else:
                lbl.grid(row=row, column=i, padx=3, sticky="ew")

    def apply_test_row(self, test: Test, row: int) -> None:
        """Creates a row for the test and grids it."""
        cols: List[tk.Widget] = []
        vcmd = self.register(self.update_score)
        # col 0 - name
        cols.append(ttk.Label(self, textvariable=test.name))

        # col 2 - duration
        duration = round(
            len(test.readings) * self.project.interval_seconds.get() / 60, 2
        )
        cols.append(
            ttk.Label(
                self,
                text=f"{duration:.2f}",
                anchor="center",
            )
        )
        # col 3 - pump to score
        to_score = ttk.Combobox(
            self,
            textvariable=test.pump_to_score,
            values=["pump 1", "pump 2", "average"],
            state="readonly",
            width=7,
            validate="all",
            validatecommand=vcmd,
        )
        to_score.bind("<MouseWheel>", self.update_score)
        cols.append(to_score)
        # col 4 - obs baseline
        cols.append(
            ttk.Label(
                self, textvariable=test.observed_baseline, anchor="center", width=5
            )
        )
        # col 5 - max psi
        cols.append(
            ttk.Label(self, textvariable=test.max_psi, anchor="center", width=7)
        )
        # col 6 - clarity
        cols.append(
            ttk.Label(
                self,
                textvariable=test.clarity,
                anchor="center",
            )
        )
        # col 7 - notes
        cols.append(ttk.Entry(self, textvariable=test.notes, width=25))
        # col 8 - result
        cols.append(ttk.Label(self, textvariable=test.result, width=5, anchor="center"))
        # col 9 - include on report
        cols.append(
            ttk.Checkbutton(
                self,
                variable=test.include_on_report,
                command=self.update_score,
            )
        )
        # col 10 - delete
        cols.append(
            ttk.Button(
                self,
                command=lambda: self.remove_from_project(test),  # noqa: E731
                text="Delete",
                width=7,
            )
        )

        for i, col in enumerate(cols):
            if i == 0:  # left align the name col
                col.grid(row=row, column=i, padx=1, pady=1, sticky="w")
            elif i == 7:  # make the notes col stretch
                col.grid(row=row, column=i, padx=1, pady=1, sticky="ew")
            elif i == 10:  # right align the delete buttons
                col.grid(row=row, column=i, padx=(5, 0), pady=1, sticky="e")
            else:
                col.grid(row=row, column=i, padx=1, pady=1)

    def sort_tests(self) -> None:
        """Sort through the project, populating the lists of blanks and trials."""
        self.blanks.clear()
        self.trials.clear()

        for test in self.project.tests:
            if test.is_blank.get():
                self.blanks.append(test)
            else:
                self.trials.append(test)

    def remove_from_project(self, test: Test) -> None:
        """Removes a Test from the parent Project, then rebuilds the UI."""
        msg = (
            "You are about to delete {} from {}.\n"
            "This will become permanent once you save the project.\n"
            "Do you wish to continue?"
        ).format(test.name.get(), self.project.name.get())
        remove = messagebox.askyesno("Delete test", msg)
        if remove and test in self.project.tests:
            self.project.tests.remove(test)
            self.update_score()
            self.build()

    def update_score(self, *args) -> True:
        """Calls score from a validation callback. Doesn't check anything."""
        # prevents a race condition when setting the score
        self.after(0, score(self.project, self.eval_window.log_text))
        if isinstance(self.eval_window.plot_view, EvaluationPlotView):
            self.after(0, self.eval_window.plot_view.update_plot)
        return True



"""This module defines functions that deal with program configuration."""

from __future__ import annotations

import os
from logging import getLogger
from pathlib import Path
from typing import Union

from appdirs import user_config_dir
from tomlkit import comment, document, dumps, loads, table

import scalewiz

LOGGER = getLogger("scalewiz.config")

CONFIG_DIR = Path(user_config_dir("ScaleWiz", "teauxfu"))
CONFIG_FILE = Path(CONFIG_DIR, "config.toml")


def ensure_config() -> None:
    """Makes a config directory if one doesn't exist.

    Args:
        directory (str): path for the config directory
    """
    # make sure we have a place to store data
    if not CONFIG_DIR.is_dir():
        LOGGER.info("No config directory found. Making one now at %s", CONFIG_DIR)
        CONFIG_DIR.mkdir(parents=True)
    # make sure the file exists and isn's empty
    if not CONFIG_FILE.is_file() or os.stat(CONFIG_FILE).st_size == 0:
        LOGGER.info(
            "No config file found in %s. Making one now at %s", CONFIG_DIR, CONFIG_FILE
        )
        init_config()


def init_config() -> None:
    """Ensures a user config dir and file."""
    doc = generate_default()
    # write to file
    with CONFIG_FILE.open("w") as file:
        file.write(dumps(doc))
    # report
    if CONFIG_FILE.is_file():
        LOGGER.info("Successfully built a new config file at %s", CONFIG_FILE)
    else:
        LOGGER.warning("Failed to init a config file at %s", CONFIG_FILE)


def generate_default() -> document:
    """Generates the default TOML doc."""
    # make the toml
    doc = document()
    # orient the user
    doc.add(comment("This is the configuration file for ScaleWiz"))
    doc.add(
        comment(
            "This is a TOML document, "
            "made according to the spec at https://toml.io/en/"
        )
    )
    doc.add(
        comment(
            "If it behaves unexpectedly, try using "
            "https://www.toml-lint.com/ to check your edits"
        )
    )
    doc.add(
        comment(
            "You may delete this file to generate a new one "
            "the next time you run ScaleWiz"
        )
    )
    doc.add(
        comment(
            "If valid TOML is behaving unexpectedly in ScaleWiz, "
            "please open an issue at https://github.com/teauxfu/scalewiz/issues"
        )
    )

    # these will get updated between user sessions
    recents = table()
    recents["analyst"] = "teauxfu"
    recents["project"] = ""
    doc["recents"] = recents
    doc["recents"].comment("these will get updated between user sessions")

    # general defaults
    params = table()

    params["baseline"] = 0
    params["baseline"].comment("psi, a positive integer")

    params["flowrate"] = 0.01
    params["flowrate"].comment("mL/min, a float => 0.01")

    params["output_format"] = "CSV"
    params["output_format"].comment('choose from ("CSV", "JSON")')

    params["pressure_limit"] = 1
    params["pressure_limit"].comment("psi, a positive integer")

    params["reading_interval"] = 1.0
    params["reading_interval"].comment("seconds between readings, a positive float")

    params["test_temperature"] = 1.0
    params["test_temperature"].comment("test temperature in °F, a positive float")

    params["time_limit"] = 1.0
    params["time_limit"].comment("minutes, a positive float")

    params["uptake_time"] = 1.0
    params["uptake_time"].comment(
        "seconds between pumps starting and data collection, a float => 0.0"
    )

    doc["defaults"] = params
    doc["defaults"].comment("these will get used when making new projects")
    # all done
    return doc


def open_config() -> None:
    """Opens the config file."""
    ensure_config()
    if CONFIG_FILE.is_file():
        os.startfile(CONFIG_FILE)


def get_config() -> dict[str, Union[float, int, str]]:
    """Returns the current configuration as a dict."""
    ensure_config()
    with CONFIG_FILE.open("r") as file:
        config = loads(file.read())
    return config


def update_config(table: str, key: str, value: Union[float, int, str]) -> None:
    """Update the config with the passed values.

    Args:
        table (str): table to update (expects "recents" or "defaults")
        key (str): the key to update
        value (Union[float, int, str]): the new value of `key`
    """
    ensure_config()
    doc = loads(CONFIG_FILE.open("r").read())
    if table in doc.keys() and key in doc[table].keys():
        doc[table][key] = value
        CONFIG_FILE.write_text(dumps(doc))
        LOGGER.info("Updated %s.%s to %s", table, key, value)
        scalewiz.CONFIG = get_config()
    else:
        LOGGER.info("Failed to update %s.%s to %s", table, key, value)



"""Model object for a project. Provides a JSON/Tk mapping."""

from __future__ import annotations

import json
import logging
import tkinter as tk
from pathlib import Path
from typing import TYPE_CHECKING

from scalewiz import CONFIG
from scalewiz.helpers.configuration import update_config
from scalewiz.helpers.sort_nicely import sort_nicely
from scalewiz.models.test import Test

if TYPE_CHECKING:
    from typing import List

LOGGER = logging.getLogger("scalewiz")


class Project:
    """Model object for a project. Provides a JSON/tkVar mapping."""

    # pylint: disable=too-many-instance-attributes

    def __init__(self) -> None:
        self.tests: List[Test] = []
        # experiment parameters that affect score
        self.baseline = tk.IntVar()
        self.limit_minutes = tk.DoubleVar()
        self.limit_psi = tk.IntVar()
        self.interval_seconds = tk.DoubleVar()
        self.flowrate = tk.DoubleVar()
        self.uptake_seconds = tk.DoubleVar()
        # report stuff
        self.output_format = tk.StringVar()
        # metadata for reporting
        self.customer = tk.StringVar()
        self.submitted_by = tk.StringVar()
        self.client = tk.StringVar()
        self.field = tk.StringVar()
        self.sample = tk.StringVar()
        self.sample_date = tk.StringVar()
        self.received_date = tk.StringVar()
        self.completed_date = tk.StringVar()
        self.name = tk.StringVar()  # identifier for the project
        self.analyst = tk.StringVar()
        self.numbers = tk.StringVar()
        self.path = tk.StringVar()  # path to the project's JSON file
        self.notes = tk.StringVar()
        self.bicarbs = tk.DoubleVar()
        self.bicarbs_increased = tk.BooleanVar()
        self.calcium = tk.DoubleVar()
        self.chlorides = tk.DoubleVar()
        self.temperature = tk.DoubleVar()  # the test temperature
        self.plot = tk.StringVar()  # path to plot local file
        self.set_defaults()  # get default values from the config
        self.add_traces()  # these need to be cleaned up later

    def set_defaults(self) -> None:
        """Sets project parameters to the defaults read from the config file."""
        defaults = CONFIG["defaults"]
        # make sure we are seeing reasonable values
        for key, value in defaults.items():
            if not isinstance(value, str) and value < 0:
                defaults[key] = value * (-1)
        # apply values
        self.baseline.set(defaults["baseline"])
        self.interval_seconds.set(defaults["reading_interval"])
        self.limit_minutes.set(defaults["time_limit"])
        self.limit_psi.set(defaults["pressure_limit"])
        self.output_format.set(defaults["output_format"])
        self.temperature.set(defaults["test_temperature"])
        self.flowrate.set(defaults["flowrate"])
        self.uptake_seconds.set(defaults["uptake_time"])
        # this must never be <= 0
        if self.interval_seconds.get() <= 0:
            self.interval_seconds.set(1)
        self.analyst.set(CONFIG["recents"]["analyst"])

    def add_traces(self) -> None:
        """Adds tkVar traces where needed. Must be cleaned up with remove_traces."""
        self.customer.trace_add("write", self.update_proj_name)
        self.client.trace_add("write", self.update_proj_name)
        self.field.trace_add("write", self.update_proj_name)
        self.sample.trace_add("write", self.update_proj_name)

    def dump_json(self, path: str = None) -> None:
        """Dump a JSON representation of the Project at the passed path."""
        if path is None:
            path = Path(self.path.get())

        blanks = {}
        trials = {}
        for test in self.tests:
            label = test.label.get().lower()
            while label in blanks or label in trials:  # make sure we don't overwrite
                label = "".join((label, " - copy"))
            if test.is_blank.get():
                blanks[label] = test
            else:
                trials[label] = test

        blank_labels = sort_nicely(list(blanks.keys()))
        trial_labels = sort_nicely(list(trials.keys()))

        tests = []
        for label in blank_labels:
            tests.append(blanks.pop(label))
        for label in trial_labels:
            tests.append(trials.pop(label))

        self.tests.clear()
        self.tests = [test for test in tests]

        this = {
            "info": {
                "customer": self.customer.get(),
                "submittedBy": self.submitted_by.get(),
                "productionCo": self.client.get(),
                "field": self.field.get(),
                "sample": self.sample.get(),
                "sampleDate": self.sample_date.get(),
                "recDate": self.received_date.get(),
                "compDate": self.completed_date.get(),
                "name": self.name.get(),
                "analyst": self.analyst.get(),
                "numbers": self.numbers.get(),
                "path": str(Path(self.path.get()).resolve()),
                "notes": self.notes.get(),
            },
            "params": {
                "bicarbonates": self.bicarbs.get(),
                "bicarbsIncreased": self.bicarbs_increased.get(),
                "calcium": self.calcium.get(),
                "chlorides": self.chlorides.get(),
                "baseline": self.baseline.get(),
                "temperature": self.temperature.get(),
                "limitPSI": self.limit_psi.get(),
                "limitMin": self.limit_minutes.get(),
                "interval": self.interval_seconds.get(),
                "flowrate": self.flowrate.get(),
                "uptake": self.uptake_seconds.get(),
            },
            "tests": [test.to_dict() for test in self.tests],
            "outputFormat": self.output_format.get(),
            "plot": str(Path(self.plot.get()).resolve()),
        }
        try:
            with Path(path).open("w") as file:
                json.dump(this, file, indent=4)
            LOGGER.info("Saved %s to %s", self.name.get(), path)
            update_config("recents", "analyst", self.analyst.get())
            update_config("recents", "project", str(Path(self.path.get()).resolve()))
        except Exception as err:
            LOGGER.exception(err)

    def load_json(self, path: str) -> None:
        """Return a Project from a passed path to a JSON dump."""
        path = Path(path).resolve()
        if path.is_file():
            LOGGER.info("Loading from %s", path)
            with path.open("r") as file:
                obj = json.load(file)

        # we expect the data files to be shared over Dropbox, etc.
        if str(path) != obj["info"]["path"]:
            LOGGER.warning(
                "Opened a Project whose actual path didn't match its path property"
            )
            obj["info"]["path"] = str(path)

        info: dict = obj["info"]
        self.customer.set(info["customer"])
        self.submitted_by.set(info["submittedBy"])
        self.client.set(info["productionCo"])
        self.field.set(info["field"])
        self.sample.set(info["sample"])
        self.sample_date.set(info["sampleDate"])
        self.received_date.set(info["recDate"])
        self.completed_date.set(info["compDate"])
        self.name.set(info["name"])
        self.numbers.set(info["numbers"])
        self.analyst.set(info["analyst"])
        self.path.set(str(Path(info["path"]).resolve()))
        self.notes.set(info["notes"])

        defaults = CONFIG["defaults"]
        params: dict = obj["params"]
        self.bicarbs.set(params.get("bicarbonates", 0))
        self.bicarbs_increased.set(params.get("bicarbsIncreased", False))
        self.calcium.set(params.get("calcium", 0))
        self.chlorides.set(params.get("chlorides", 0))
        self.baseline.set(params.get("baseline", defaults["baseline"]))
        self.temperature.set(params.get("temperature", defaults["test_temperature"]))
        self.limit_psi.set(params.get("limitPSI", defaults["pressure_limit"]))
        self.limit_minutes.set(params.get("limitMin", defaults["time_limit"]))
        self.interval_seconds.set(params.get("interval", defaults["reading_interval"]))
        self.flowrate.set(params.get("flowrate", defaults["flowrate"]))
        self.uptake_seconds.set(params.get("uptake", defaults["uptake_time"]))
        self.output_format.set(obj.get("outputFormat", defaults["output_format"]))

        self.plot.set(obj["plot"])

        self.tests.clear()
        for entry in obj["tests"]:
            test = Test(data=entry)
            self.tests.append(test)

    def remove_traces(self) -> None:
        """Remove tkVar traces to allow the GC to do its thing."""
        variables = (self.customer, self.client, self.field, self.sample)
        for var in variables:
            try:
                var.trace_remove("write", var.trace_info()[0][1])
            except IndexError:  # sometimes this spaghets when loading empty projects...
                pass
        for test in self.tests:
            test.remove_traces()

    def update_proj_name(self, *args) -> None:
        """Constructs a default name for the Project."""
        # extra unused args are passed in by tkinter
        if self.client.get() != "":
            name = self.client.get().strip()
        else:
            name = self.customer.get().strip()
        if self.field.get() != "":
            name = f"{name} - {self.field.get().strip()}"
        if self.sample.get() != "":
            name = f"{name} ({self.sample.get().strip()})"
        self.name.set(name)



"""Sets an icon on the current Toplevel."""
from __future__ import annotations

import logging
import os
import tkinter as tk
from pathlib import Path

from scalewiz.helpers.get_resource import get_resource

LOGGER = logging.getLogger("scalewiz")


def set_icon(widget: tk.Widget) -> None:
    """Sets an icon on the current Toplevel."""
    # set the Toplevel's icon
    try:  # this makes me nervous, but whatever
        icon_path = Path(get_resource(r"../components/icon.ico")).resolve()
        if icon_path.is_file():
            widget.winfo_toplevel().wm_iconbitmap(icon_path)
        # for windows, set the taskbar icon
        if "nt" in os.name:
            import ctypes

            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID("scalewiz")
    except FileNotFoundError:
        LOGGER.error("Failed to set the icon")



"""A plot view to be displayed in the Evaluation Window."""

from __future__ import annotations

import tkinter as tk
from logging import Logger, getLogger
from tkinter import Canvas, ttk
from tkinter.font import Font
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.figure import Figure, SubplotParams
from matplotlib.ticker import MultipleLocator

if TYPE_CHECKING:
    from matplotlib.axis import Axis

    from scalewiz.models.project import Project

LOGGER: Logger = getLogger("scalewiz")


class EvaluationPlotView(ttk.Frame):
    """A widget for selecting devices."""

    def __init__(self, parent: ttk.Notebook, project: Project) -> None:
        super().__init__(parent)
        self.parent: ttk.Notebook = parent
        self.project: Project = project
        self.fig: Figure = None
        self.canvas: Canvas = None
        self.axis: Axis = None
        self.plot_frame: ttk.Frame = None
        self.build()

    def build(self) -> None:
        """Builds the UI."""
        if not self.winfo_exists():
            return

        if isinstance(self.fig, Figure):
            plt.close(self.fig)

        for child in self.winfo_children():
            if child.winfo_exists():
                child.destroy()

        self.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(0, weight=1)
        self.grid_columnconfigure(1, weight=1)

        label_frame = ttk.Frame(self)
        bold_font = Font(family="Arial", weight="bold", size=10)
        label_lbl = tk.Label(
            label_frame, text="Label", font=bold_font, width=20, anchor="center"
        )
        label_lbl.grid(row=0, column=0, sticky="ew")

        vcmd = self.register(self.update_plot)

        tests_on_report = []
        for test in self.project.tests:
            if test.include_on_report.get():
                tests_on_report.append(test)

        for i, test in enumerate(tests_on_report):
            label_ent = ttk.Entry(
                label_frame,
                textvariable=test.label,
                validate="focusout",
                validatecommand=vcmd,
                width=25,
            )
            label_ent.grid(row=i + 1, column=0, sticky="ew", pady=2)

        label_frame.grid(row=0, column=1, sticky="nsew")

        self.plot_frame = ttk.Frame(self)
        self.fig, self.axis = plt.subplots(
            figsize=(7.5, 4),
            dpi=100,
            subplotpars=SubplotParams(
                wspace=0, hspace=0, left=0.1, right=0.9, top=0.95
            ),
        )
        self.fig.patch.set_facecolor("#FAFAFA")
        self.canvas = FigureCanvasTkAgg(self.fig, master=self.plot_frame)
        self.canvas.get_tk_widget().pack(fill="both", expand=True)
        with plt.style.context("bmh"):
            self.axis.grid(color="darkgrey", alpha=0.65, linestyle="-")
            self.axis.set_facecolor("w")  # white

            # plot blanks
            for test in tests_on_report:
                if test.is_blank.get():
                    elapsed = []
                    for reading in test.readings:
                        elapsed.append(reading.elapsedMin)
                    self.axis.plot(
                        elapsed,
                        test.get_readings(),
                        label=test.label.get(),
                        linestyle=("-."),
                    )
            # then plot trials
            for test in tests_on_report:
                if not test.is_blank.get():
                    elapsed = []
                    for reading in test.readings:
                        elapsed.append(reading.elapsedMin)
                    self.axis.plot(elapsed, test.get_readings(), label=test.label.get())

            self.axis.set_xlabel("Time (min)")
            self.axis.set_ylabel("Pressure (psi)")
            self.axis.set_ylim(top=self.project.limit_psi.get())
            self.axis.yaxis.set_major_locator(MultipleLocator(100))
            self.axis.set_xlim((0, self.project.limit_minutes.get()))
            self.axis.legend(loc="best")
            self.axis.margins(0)

        self.plot_frame.grid(row=0, column=0, sticky="n")

    def update_plot(self) -> True:
        """Rebuilds the plot."""
        self.after_idle(self.build)
        return True



"""A function for exporting a representation of a Project as CSV."""

from __future__ import annotations

import json
import logging
from pathlib import Path
from typing import Tuple

from pandas import DataFrame

from scalewiz.models.project import Project

LOGGER = logging.getLogger("scalewiz")


def export(project: Project) -> Tuple[int, Path]:
    """Generates a report for a Project in a flattened CSV format (or ugly JSON)."""
    LOGGER.info("Beginning export of %s", project.name.get())

    output_dict = {
        "customer": project.customer.get(),
        "submitteBy": project.submitted_by.get(),
        "productionCompany": project.client.get(),
        "field": project.field.get(),
        "samplePoint": project.sample.get(),
        "analysisNumbers": project.numbers.get(),
        "dateSampled": project.sample_date.get(),
        "dateReceived": project.received_date.get(),
        "dateCompleted": project.completed_date.get(),
        "testTempF": project.temperature.get(),
        "baselinePsi": project.baseline.get(),
        "bicarbs": project.bicarbs.get(),
        "bicarbsIncreased": project.bicarbs_increased.get(),
        "calcium": project.calcium.get(),
        "chlorides": project.chlorides.get(),
        "timeLimitMin": project.limit_minutes.get(),
        "limitPsi": project.limit_psi.get(),
        "readingIntervalSecs": project.interval_seconds.get(),
        "name": [],
        "isBlank": [],
        "chemical": [],
        "rate": [],
        "duration": [],
        "maxPsi": [],
        "result": [],
        "clarity": [],
        "plotPath": project.plot.get(),
    }
    # filter the blanks and trials to sort them
    blanks = [
        test
        for test in project.tests
        if test.include_on_report.get() and test.is_blank.get()
    ]
    trials = [
        test
        for test in project.tests
        if test.include_on_report.get() and not test.is_blank.get()
    ]
    tests = blanks + trials
    # we use lists here instead of sets since sets aren't JSON serializable
    output_dict["name"] = [test.name.get() for test in tests]
    output_dict["isBlank"] = [test.is_blank.get() for test in tests]
    output_dict["chemical"] = [test.chemical.get() for test in tests]
    output_dict["rate"] = [test.rate.get() for test in tests]
    output_dict["duration"] = [
        round(len(test.readings) * project.interval_seconds.get() / 60, 2)
        for test in tests
    ]
    output_dict["maxPsi"] = [test.max_psi.get() for test in tests]
    output_dict["result"] = [test.result.get() for test in tests]
    output_dict["clarity"] = [test.clarity.get() for test in tests]

    fmt = project.output_format.get()
    out = f"{project.numbers.get().replace(' ', '')} {project.name.get()}"
    out = f"{out} - CaCO3 Scale Block Analysis.{fmt}".strip()
    out = Path(Path(project.path.get()).parent).joinpath(out).resolve()

    with out.open("w") as output:
        if fmt == "CSV":
            data = DataFrame.from_dict(output_dict)
            data.to_csv(out, encoding="utf-8")
        elif fmt == "JSON":
            json.dump(output_dict, output, indent=4)

    LOGGER.info(
        "Finished export of %s as %s",
        project.name.get(),
        project.output_format.get(),
    )

    if out.is_file():
        return 0, out
    else:
        return 1, out



from __future__ import annotations

import tkinter as tk
from logging import Logger, getLogger
from queue import Empty
from tkinter import ttk
from tkinter.scrolledtext import ScrolledText
from typing import TYPE_CHECKING

if TYPE_CHECKING:

    from scalewiz.models.test_handler import TestHandler


LOGGER: Logger = getLogger("scalewiz")


class TestControls(ttk.Frame):
    """A widget for selecting devices."""

    def __init__(self, parent: tk.Widget, handler: TestHandler) -> None:
        super().__init__(parent)
        self.handler: TestHandler = handler
        self.interval: int = round(handler.project.interval_seconds.get() * 1000)
        self.build()

    def build(self) -> None:
        self.grid_columnconfigure(0, weight=1)
        self.grid_columnconfigure(1, weight=1)
        # row 0 col 0
        start_btn = ttk.Button(self)
        if self.handler.is_done and not self.handler.is_running:
            start_btn.configure(text="New", command=self.handler.new_test)
        elif self.handler.is_running and not self.handler.is_done:
            start_btn.configure(
                text="New", command=self.handler.new_test, state="disabled"
            )
        else:
            start_btn.configure(text="Start", command=self.handler.start_test)
        start_btn.grid(row=0, column=0, sticky="ew")
        # row 0 col 1
        if self.handler.is_running:
            state = "normal"
        else:
            state = "disabled"
        stop_btn = ttk.Button(
            self, text="Stop", command=self.handler.request_stop, state=state
        )
        stop_btn.grid(row=0, column=1, sticky="ew")
        progressbar = ttk.Progressbar(self, variable=self.handler.progress, maximum=100)
        progressbar.grid(row=1, column=0, columnspan=2, sticky="ew")
        # row 1 col 0:1
        self.log_text = ScrolledText(
            self, background="white", height=5, width=44, state="disabled"
        )
        self.log_text.grid(row=2, column=0, columnspan=2, sticky="ew")
        # enter polling loop
        self.after(0, self.poll_log_queue)

    def poll_log_queue(self) -> None:
        """Checks on an interval if there is a new message in the queue to display."""
        while True:
            try:
                record = self.handler.log_queue.get(block=False)
            except Empty:
                break
            else:
                self.display(record)
        self.after(self.interval, self.poll_log_queue)

    def display(self, msg: str) -> None:
        """Displays a message in the log."""
        self.log_text.configure(state="normal")
        self.log_text.insert("end", "".join((msg, "\n")))
        self.log_text.configure(state="disabled")
        self.log_text.yview("end")  # scroll to bottom



"""Editor for Project metadata."""

from __future__ import annotations

from tkinter import ttk
from typing import TYPE_CHECKING

import tkcalendar as tkcal

if TYPE_CHECKING:
    from scalewiz.models.project import Project


def render(lbl: ttk.Label, ent: ttk.Entry, row: int) -> None:
    """Grids a label and entry on the passed row."""
    lbl.grid(row=row, column=0, sticky="e")
    ent.grid(row=row, column=1, sticky="ew", pady=1)


class ProjectInfo(ttk.Frame):
    """Editor for Project metadata."""

    def __init__(self, parent: ttk.Frame, project: Project) -> None:
        super().__init__(parent)
        self.grid_columnconfigure(1, weight=1)

        # row 0 -----------------------------------------------------------------------
        lbl = ttk.Label(self, text="Customer:")
        ent = ttk.Entry(self, textvariable=project.customer)
        lbl.grid(row=0, column=0, sticky="e")
        render(lbl, ent, 0)

        # row 1 -----------------------------------------------------------------------
        lbl = ttk.Label(self, text="Submitted by:")
        ent = ttk.Entry(self, textvariable=project.submitted_by)
        render(lbl, ent, 1)

        # row 2 -----------------------------------------------------------------------
        lbl = ttk.Label(self, text="Production company:")
        ent = ttk.Entry(self, textvariable=project.client)
        render(lbl, ent, 2)

        # row 3 -----------------------------------------------------------------------
        lbl = ttk.Label(self, text="Field:")
        ent = ttk.Entry(self, textvariable=project.field)
        render(lbl, ent, 3)

        # row 4 -----------------------------------------------------------------------
        lbl = ttk.Label(self, text="Sample point:")
        ent = ttk.Entry(self, textvariable=project.sample)
        render(lbl, ent, 4)

        # row 5 -----------------------------------------------------------------------
        lbl = ttk.Label(self, text="Project name:")
        ent = ttk.Entry(self, textvariable=project.name)
        render(lbl, ent, 5)

        # row 6 -----------------------------------------------------------------------
        lbl = ttk.Label(self, text="Date sampled:")
        lbl.bind("<Button-1>", lambda _: project.sample_date.set(""))
        # this is to refresh the entry later (it inits with today's date)
        current_value = project.sample_date.get()
        ent = tkcal.DateEntry(
            self, textvariable=project.sample_date, date_pattern="mm/dd/yyyy"
        )
        lbl.bind("<Button-1>", lambda _: project.sample_date.set(""))
        project.sample_date.set(current_value)
        render(lbl, ent, 6)

        # row 7 -----------------------------------------------------------------------
        lbl = ttk.Label(self, text="Date received:")
        lbl.bind("<Button-1>", lambda _: project.received_date.set(""))
        current_value = project.received_date.get()
        ent = tkcal.DateEntry(
            self,
            textvariable=project.received_date,
            date_pattern="mm/dd/yyyy",
        )
        project.received_date.set(current_value)
        render(lbl, ent, 7)

        # row 8 -----------------------------------------------------------------------
        lbl = ttk.Label(self, text="Date completed:")
        lbl.bind("<Button-1>", lambda _: project.completed_date.set(""))
        current_value = project.completed_date.get()
        ent = tkcal.DateEntry(
            self,
            textvariable=project.completed_date,
            date_pattern="mm/dd/yyyy",
        )
        project.completed_date.set(current_value)
        render(lbl, ent, 8)

        # row 9 -----------------------------------------------------------------------
        lbl = ttk.Label(self, text="Analyst:")
        ent = ttk.Entry(self, textvariable=project.analyst)
        render(lbl, ent, 9)

        # row 10 ----------------------------------------------------------------------
        lbl = ttk.Label(self, text="Analysis number(s):")
        ent = ttk.Entry(self, textvariable=project.numbers)
        render(lbl, ent, 10)

        # row 11 ----------------------------------------------------------------------
        lbl = ttk.Label(self, text="Notes:")
        ent = ttk.Entry(self, textvariable=project.notes)
        render(lbl, ent, 11)

        # row 12 ----------------------------------------------------------------------
        pathLbl = ttk.Label(self, text="File path:")
        pathEnt = ttk.Label(self, textvariable=parent.editor_project.path)
        render(pathLbl, pathEnt, 12)



"""Sort the given list in the way that humans expect."""

# thanks https://blog.codinghorror.com/sorting-for-humans-natural-sort-order/
# but modifying in-place seems rude
# http://www.compciv.org/guides/python/fundamentals/lists-mutability/
import re


def sort_nicely(things: list) -> list:
    """Sort the given list in the way that humans expect."""

    new_list = things.copy()
    convert = lambda text: int(text) if text.isdigit() else text  # noqa: E731
    alnum_key = lambda key: [  # noqa: E731
        convert(c) for c in re.split("([0-9]+)", key)
    ]
    new_list.sort(key=alnum_key)
    return new_list



"""Core object for the application. Used for defining application-scope settings."""

import logging
import os
from importlib.metadata import version
from logging.handlers import QueueHandler
from queue import Queue
from tkinter import font, ttk

from scalewiz.components.scalewiz_log_window import LogWindow
from scalewiz.components.scalewiz_main_frame import MainFrame
from scalewiz.helpers.set_icon import set_icon


class ScaleWiz(ttk.Frame):
    """Core object for the application.

    Used to define widget styles and set up logging.
    """

    def __init__(self, parent) -> None:
        super().__init__(parent)
        # set UI
        # icon / version
        set_icon(parent)
        parent.title(f"ScaleWiz {version('scalewiz')}")
        parent.resizable(0, 0)  # apparently this is a bad practice...
        # font 🔠
        default_font = font.nametofont("TkDefaultFont")
        default_font.configure(family="Arial")
        parent.option_add("*Font", "TkDefaultFont")
        bold_font = font.Font(family="Helvetica", weight="bold")
        ttk.Style().configure("TNotebook.Tab", font=bold_font)
        # widget backgrounds / themes 🎨
        parent.tk_setPalette(background="#FAFAFA")
        for aspect in ("TLabel", "TFrame", "TRadiobutton", "TCheckbutton", "TNotebook"):
            ttk.Style().configure(aspect, background="#FAFAFA")
        # configure logging functionality
        self.log_queue = Queue()
        queue_handler = QueueHandler(self.log_queue)
        # this is for inspecting the multithreading
        fmt = "%(asctime)s - %(thread)d - %(levelname)s - %(name)s - %(message)s"
        # fmt = "%(asctime)s - %(levelname)s - %(name)s - %(message)s"
        date_fmt = "%Y-%m-%d %H:%M:%S"
        formatter = logging.Formatter(
            fmt,
            date_fmt,
        )
        logging.basicConfig(level=logging.DEBUG)  # applies to the root logger instance
        queue_handler.setFormatter(formatter)
        queue_handler.setLevel(logging.INFO)
        logger = logging.getLogger("scalewiz")
        logger.addHandler(queue_handler)
        # holding a ref to the toplevel for the menubar to find
        self.log_window = LogWindow(self)
        logger.info("Starting in %s", os.getcwd())
        self.log_window.withdraw()  # 🏌️‍♀️👋
        MainFrame(self).grid()



"""Editor for Project params."""

from __future__ import annotations

from tkinter import ttk
from typing import TYPE_CHECKING

from scalewiz.helpers.validation import can_be_float, can_be_pos_float, can_be_pos_int

if TYPE_CHECKING:
    from scalewiz.models.project import Project


def render(lbl: ttk.Label, ent: ttk.Entry, row: int) -> None:
    """Grids a label and entry on the passed row."""
    lbl.grid(row=row, column=0, sticky="e")
    ent.grid(row=row, column=1, sticky="ew", pady=1)


class ProjectParams(ttk.Frame):
    """A form for mutating experiment-relevant attributes of the Project."""

    def __init__(self, parent: ttk.Frame, project: Project) -> None:
        super().__init__(parent)
        # validation commands to ensure numeric inputs

        is_pos_int = self.register(lambda s: can_be_pos_int(s))
        is_float = self.register(lambda s: can_be_float(s))
        is_pos_float = self.register(lambda s: can_be_pos_float(s))
        # https://anzeljg.github.io/rin2/book2/2405/docs/tkinter/entry-validation.html

        self.grid_columnconfigure(1, weight=1)
        # row 0 ------------------------------------------------------------------------
        lbl = ttk.Label(self, text="Bicarbonates (mg/L):")
        ent = ttk.Spinbox(
            self,
            textvariable=project.bicarbs,
            from_=0,
            to=999999,
            validate="key",
            validatecommand=(is_float, "%P"),
        )
        render(lbl, ent, 0)

        # row 1 ------------------------------------------------------------------------
        lbl = ttk.Label(self, text="Bicarbonates increased:")
        ent = ttk.Frame(self)
        ent.grid_columnconfigure(0, weight=1)
        ent.grid_columnconfigure(1, weight=1)
        ttk.Radiobutton(
            ent,
            text="Yes",
            variable=project.bicarbs_increased,
            value=True,
        ).grid(row=0, column=0)
        ttk.Radiobutton(
            ent,
            text="No",
            variable=project.bicarbs_increased,
            value=False,
        ).grid(row=0, column=1)
        render(lbl, ent, 1)

        # row 2 ------------------------------------------------------------------------
        lbl = ttk.Label(self, text="Calcium (mg/L):")
        ent = ttk.Spinbox(
            self,
            textvariable=project.calcium,
            from_=0,
            to=999999,
            validate="key",
            validatecommand=(is_float, "%P"),
        )
        render(lbl, ent, 2)

        # row 3 ------------------------------------------------------------------------
        lbl = ttk.Label(self, text="Chlorides (mg/L):")
        ent = ttk.Spinbox(
            self,
            textvariable=project.chlorides,
            from_=0,
            to=999999,
            validate="key",
            validatecommand=(is_float, "%P"),
        )
        render(lbl, ent, 3)

        # row 4 ------------------------------------------------------------------------
        lbl = ttk.Label(self, text="Test temperature (°F):")
        ent = ttk.Spinbox(
            self,
            textvariable=project.temperature,
            from_=0,
            to=9999,
            validate="key",
            validatecommand=(is_pos_float, "%P"),
        )
        render(lbl, ent, 4)

        # row 5 ------------------------------------------------------------------------
        lbl = ttk.Label(self, text="Baseline pressure (PSI):")
        ent = ttk.Spinbox(
            self,
            textvariable=project.baseline,
            from_=0,
            to=9999,
            validate="key",
            validatecommand=(is_pos_int, "%P"),
        )
        render(lbl, ent, 5)

        # row 6 ------------------------------------------------------------------------
        lbl = ttk.Label(self, text="Limiting pressure (PSI):")
        ent = ttk.Spinbox(
            self,
            textvariable=project.limit_psi,
            from_=0,
            to=9999,
            validate="key",
            validatecommand=(is_pos_int, "%P"),
        )
        render(lbl, ent, 6)

        # row 7 ------------------------------------------------------------------------
        lbl = ttk.Label(self, text="Time limit (min.):")
        ent = ttk.Spinbox(
            self,
            textvariable=project.limit_minutes,
            from_=0,
            to=9999,
            validate="key",
            validatecommand=(is_pos_float, "%P"),
        )
        render(lbl, ent, 7)

        # row 8 ------------------------------------------------------------------------
        lbl = ttk.Label(self, text="Reading interval (s):")
        ent = ttk.Spinbox(
            self,
            textvariable=project.interval_seconds,
            from_=1,
            to=9999,
            validate="key",
            validatecommand=(is_pos_float, "%P"),
        )
        render(lbl, ent, 8)

        # row 9 ------------------------------------------------------------------------
        lbl = ttk.Label(self, text="Flowrate (mL/min):")
        ent = ttk.Spinbox(
            self,
            textvariable=project.flowrate,
            from_=1,
            to=10,
            validate="key",
            validatecommand=(is_pos_float, "%P"),
        )
        render(lbl, ent, 9)

        # row 10 -----------------------------------------------------------------------
        lbl = ttk.Label(self, text="Uptake time (s):")
        ent = ttk.Spinbox(
            self,
            textvariable=project.uptake_seconds,
            from_=0,
            to=9999,
            validate="key",
            validatecommand=(is_float, "%P"),
        )
        render(lbl, ent, 10)



"""A Toplevel with a ScrolledText. Displays messages from a Logger."""

from __future__ import annotations

import tkinter as tk
from queue import Empty
from tkinter.scrolledtext import ScrolledText
from typing import TYPE_CHECKING

from scalewiz.helpers.set_icon import set_icon

if TYPE_CHECKING:
    from logging import LogRecord

    from scalewiz.components.scalewiz import ScaleWiz


# thanks https://github.com/beenje/tkinter-logging-text-widget


class LogWindow(tk.Toplevel):
    """A Toplevel with a ScrolledText. Displays messages from a Logger."""

    def __init__(self, core: ScaleWiz) -> None:
        super().__init__()
        self.log_queue = core.log_queue
        self.title("Log Window")
        # replace the window closing behavior with withdrawing instead 🐱‍👤
        self.protocol("WM_DELETE_WINDOW", lambda: self.winfo_toplevel().withdraw())
        self.build()

    def build(self) -> None:
        """Builds the UI."""
        set_icon(self)
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)
        self.scrolled_text = ScrolledText(self, state="disabled", width=88)
        self.scrolled_text.grid(row=0, column=0, sticky="nsew")
        self.scrolled_text.tag_config("INFO", foreground="black")
        self.scrolled_text.tag_config("DEBUG", foreground="gray")
        self.scrolled_text.tag_config("WARNING", foreground="orange")
        self.scrolled_text.tag_config("ERROR", foreground="red")
        self.scrolled_text.tag_config("CRITICAL", foreground="red", underline=1)
        # start polling messages from the queue 📩
        self.after(100, self.poll_log_queue)

    def poll_log_queue(self) -> None:
        """Checks every 100ms if there is a new message in the queue to display."""
        while True:
            try:
                record = self.log_queue.get(block=False)
            except Empty:
                break
            else:
                self.display(record)
        self.after(100, self.poll_log_queue)

    def display(self, record: LogRecord) -> None:
        """Displays a message in the log."""
        msg = record.getMessage()
        self.scrolled_text.configure(state="normal")
        self.scrolled_text.insert(
            "end", "".join((msg, "\n")), record.levelname
        )  # last arg is for the tag
        self.scrolled_text.configure(state="disabled")
        self.scrolled_text.yview("end")  # scroll to bottom



"""Form for mutating Project objects."""

from __future__ import annotations

import tkinter as tk
from pathlib import Path
from tkinter import filedialog, messagebox, ttk
from typing import TYPE_CHECKING

from scalewiz.components.project_editor_info import ProjectInfo
from scalewiz.components.project_editor_params import ProjectParams
from scalewiz.components.project_editor_report import ProjectReport
from scalewiz.helpers.configuration import open_config
from scalewiz.helpers.set_icon import set_icon
from scalewiz.models.project import Project

if TYPE_CHECKING:
    from scalewiz.models.test_handler import TestHandler


class ProjectWindow(tk.Toplevel):
    """Form for mutating Project objects.

    Has a tab control widget for separating the sub-forms.
    """

    def __init__(self, handler: TestHandler) -> None:
        super().__init__()
        self.handler: TestHandler = handler
        self.editor_project: Project = Project()
        if Path(handler.project.path.get()).is_file():
            self.editor_project.load_json(handler.project.path.get())

        self.title(f"{self.handler.name}")
        set_icon(self)
        self.build()

    def build(self, reload: bool = False) -> None:
        """Destroys all child widgets, then builds the UI."""
        if reload:
            self.editor_project.remove_traces()  # clean up the old one for GC
            self.editor_project = Project()
            self.editor_project.load_json(self.handler.project.path.get())

        for child in self.winfo_children():
            self.after(0, child.destroy)

        self.winfo_toplevel().resizable(0, 0)
        self.grid_columnconfigure(0, weight=1)
        tab_control = ttk.Notebook(self)
        tab_control.grid(row=0, column=0)
        tab_control.add(ProjectInfo(self, self.editor_project), text="Project info")
        tab_control.add(
            ProjectParams(self, self.editor_project), text="Experiment parameters"
        )
        tab_control.add(
            ProjectReport(self, self.editor_project), text="Report settings"
        )

        button_frame = ttk.Frame(self)

        if self.handler.is_running:
            state = "disabled"
        else:
            state = "normal"

        ttk.Button(
            button_frame, text="Save", width=7, command=self.save, state=state
        ).grid(row=0, column=0, padx=5)
        ttk.Button(
            button_frame, text="Save as", width=7, command=self.save_as, state=state
        ).grid(row=0, column=1, padx=10)
        ttk.Button(
            button_frame, text="New", width=7, command=self.new, state=state
        ).grid(row=0, column=2, padx=5)
        ttk.Button(
            button_frame, text="Edit defaults", width=10, command=self.edit
        ).grid(row=0, column=3, padx=5)
        button_frame.grid(row=1, column=0)

    def new(self) -> None:
        """Resets the form by connecting to a new Project."""
        self.editor_project = Project()
        self.build()

    def save(self) -> None:
        """Save the current Project to file as JSON."""
        # todo don't allow saving if saving to current project - otherwise fine
        if not self.handler.is_running:
            if self.editor_project.path.get() == "":
                self.save_as()
            else:
                self.editor_project.dump_json()
                self.handler.load_project(self.editor_project.path.get())
                self.handler.rebuild_views()
        else:
            messagebox.showwarning("Can't save while a Test is running")

    def save_as(self) -> None:
        """Saves the Project to JSON using a Save As dialog."""
        file_path = filedialog.asksaveasfilename(
            title="Save Project As:",
            filetypes=[("JSON files", "*.json")],
            initialfile=f"{self.editor_project.name.get()}.json",
        )

        if file_path != "":
            # make sure it is JSON extension
            ext = file_path[-5:]
            if ext not in (".json", ".JSON"):
                file_path = f"{file_path}.json"
            self.editor_project.path.set(str(Path(file_path).resolve()))
            self.save()

    def edit(self) -> None:
        """Open the program config file."""
        open_config()



"""MenuBar for the MainWindow."""

from __future__ import annotations

import logging
import tkinter as tk
from pathlib import Path

# from time import time
from tkinter.messagebox import showinfo
from typing import TYPE_CHECKING

from scalewiz.components.evaluation_window import EvaluationWindow
from scalewiz.components.project_editor import ProjectWindow
from scalewiz.components.scalewiz_rinse_window import RinseWindow
from scalewiz.helpers.show_help import show_help

LOGGER = logging.getLogger("scalewiz")

if TYPE_CHECKING:
    from scalewiz.components.handler_view import TestHandlerView
    from scalewiz.components.scalewiz_main_frame import MainFrame


class MenuBar:
    """Menu bar to be displayed on the Main Frame."""

    def __init__(self, parent: MainFrame) -> None:
        # expecting parent to be the toplevel parent of the main frame
        self.parent = parent

        menubar = tk.Menu()
        menubar.add_command(label="Add System", command=self.parent.add_handler)
        # make project cascade
        project_menu = tk.Menu(tearoff=0)
        project_menu.add_command(label="New/Edit", command=self.spawn_editor)
        project_menu.add_command(
            label="Load existing", command=self.request_project_load
        )
        menubar.add_cascade(label="Project", menu=project_menu)
        # resume making buttons
        menubar.add_command(label="Evaluation", command=self.spawn_evaluator)
        menubar.add_command(label="Rinse", command=self.spawn_rinse)
        menubar.add_command(
            label="Log", command=self.parent.master.log_window.deiconify
        )
        menubar.add_command(label="Help", command=show_help)
        menubar.add_command(label="About", command=self.about)

        # menubar.add_command(label="Debug", command=self._debug)
        self.menubar = menubar

    def spawn_editor(self) -> None:
        """Spawn a Toplevel for editing Projects."""
        current_tab = self.parent.tab_control.select()
        widget = self.parent.nametowidget(current_tab)
        window = ProjectWindow(widget.handler)
        widget.handler.views.append(window)
        LOGGER.debug("Spawned a Project Editor window for %s", widget.handler.name)

    def spawn_evaluator(self) -> None:
        """Requests to open an evalutaion window for the currently selected Project."""
        current_tab = self.parent.tab_control.select()
        widget: TestHandlerView = self.parent.nametowidget(current_tab)
        window = EvaluationWindow(widget.handler)
        widget.handler.views.append(window)
        LOGGER.debug("Spawned an Evaluation window for %s", widget.handler.name)

    def request_project_load(self) -> None:
        """Request that the currently selected TestHandler load a Project."""
        # build a list of currently loaded projects, and pass to the handler
        currently_loaded = set()
        for tab in self.parent.tab_control.tabs():
            widget = self.parent.nametowidget(tab)
            currently_loaded.add(Path(widget.handler.project.path.get()))
        # the handler will check to make sure we don't load a project in duplicate
        current_tab = self.parent.tab_control.select()
        widget = self.parent.nametowidget(current_tab)
        widget.handler.load_project(loaded=currently_loaded)
        widget.build()

    def spawn_rinse(self) -> None:
        """Shows a RinseFrame in a new Toplevel."""
        current_tab = self.parent.tab_control.select()
        widget = self.parent.nametowidget(current_tab)
        RinseWindow(widget.handler)
        LOGGER.debug("Spawned a Rinse window for %s", widget.handler.name)

    def about(self) -> None:
        showinfo(
            "About",
            (
                "Copyright  (C) 2021  Alex Whittington\n\n"
                "This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.\n\n"  # noqa: E501
                "This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.\n\n"  # noqa: E501
                "You should have received a copy of the GNU General Public License along with this program.  If not, see <https://www.gnu.org/licenses/>."  # noqa: E501
            ),
        )

    def _debug(self) -> None:
        """Used for debugging."""
        pass
        # LOGGER.warn("DEBUGGING")

        # current_tab = self.parent.tab_control.select()
        # widget: TestHandlerView = self.parent.nametowidget(current_tab)
        # widget.handler.setup_pumps()
        # t0 = time()
        # widget.handler.pump1.pressure
        # widget.handler.pump2.pressure
        # t1 = time()
        # widget.handler.close_pumps()
        # LOGGER.warn("collected 2 pressures in %s", t1 - t0)
        # widget.handler.rebuild_views()
        # widget.bell()



"""Model object for a Test."""

from __future__ import annotations

# util
import logging
import tkinter as tk
from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import List, Tuple, Union

LOGGER = logging.getLogger("scalewiz")


@dataclass
class Reading:
    elapsedMin: float
    pump1: int
    pump2: int
    average: int


class Test:
    """Object for holding all the data associated with a Test."""

    # pylint: disable=too-many-instance-attributes

    def __init__(self, data: dict = None) -> None:
        self.is_blank = tk.BooleanVar()  # boolean for blank vs chemical trial
        self.name = tk.StringVar()  # identifier for the test
        self.chemical = tk.StringVar()  # chemical, if any, to be tested
        self.rate = tk.DoubleVar()  # the treating rate of the test
        self.label = tk.StringVar()  # how the test will be labeled on the report/plot
        self.clarity = tk.StringVar()  # the clarity of the treated water
        self.notes = tk.StringVar()  # misc notes on the experiment
        self.pump_to_score = tk.StringVar()  # which series of PSIs to use
        self.result = tk.DoubleVar()  # represents the test's performance vs the blank
        self.include_on_report = tk.BooleanVar()  # condition for scoring
        self.readings: List[Reading] = []  # list of flat reading dicts
        self.max_psi = tk.IntVar()  # the highest psi of the test
        self.observed_baseline = tk.IntVar()  # a guess at the baseline for the test
        # set defaults
        self.pump_to_score.set("pump 1")
        self.is_blank.set(True)
        self.add_traces()  # will need to clean these up later for the GC

        if isinstance(data, dict):
            self.load_json(data)

    def add_traces(self) -> None:
        """Adds tkVar traces. Need to be removed with remove_traces."""
        self.chemical.trace_add("write", self.update_test_name)
        self.rate.trace_add("write", self.update_test_name)
        self.name.trace_add("write", self.update_label)
        self.pump_to_score.trace_add("write", self.update_obs_baseline)

    def to_dict(self) -> dict[str, Union[bool, float, int, str]]:
        """Returns a dict representation of a Test."""
        self.clean_test()  # strip whitespaces from relevant fields
        # cast all readings from dataclasses to dicts
        readings = []
        for reading in self.readings:
            readings.append(
                {
                    "pump 1": reading.pump1,
                    "pump 2": reading.pump2,
                    "average": reading.average,
                    "elapsedMin": reading.elapsedMin,
                }
            )

        return {
            "name": self.name.get(),
            "isBlank": self.is_blank.get(),
            "chemical": self.chemical.get(),
            "rate": self.rate.get(),
            "reportAs": self.label.get(),
            "clarity": self.clarity.get(),
            "notes": self.notes.get(),
            "toConsider": self.pump_to_score.get(),
            "includeOnRep": self.include_on_report.get(),
            "result": self.result.get(),
            "obsBaseline": self.observed_baseline.get(),
            "readings": readings,
        }

    def load_json(self, obj: dict[str, Union[bool, float, int, str]]) -> None:
        """Load a Test with values from a JSON object."""
        self.name.set(obj["name"])
        self.is_blank.set(obj["isBlank"])
        self.chemical.set(obj["chemical"])
        self.rate.set(obj["rate"])
        self.label.set(obj["reportAs"])
        self.clarity.set(obj["clarity"])
        self.notes.set(obj["notes"])
        self.pump_to_score.set(obj["toConsider"])
        self.include_on_report.set(obj["includeOnRep"])
        self.result.set(obj["result"])
        readings = obj["readings"]
        for reading in readings:
            self.readings.append(
                Reading(
                    pump1=reading["pump 1"],
                    pump2=reading["pump 2"],
                    average=reading["average"],
                    elapsedMin=reading["elapsedMin"],
                )
            )
        self.update_obs_baseline()

    def get_readings(self) -> Tuple[int]:
        """Returns a list of the pump_to_score's pressure readings."""
        pump = self.pump_to_score.get()
        pump = pump.replace(" ", "")  # legacy accomodation for spaces in keys
        return [getattr(reading, pump) for reading in self.readings]

    def update_test_name(self, *args) -> None:
        """Makes a name by concatenating the chemical name and rate."""
        if not (self.chemical.get() == "" or self.rate.get() == 0):
            if float(self.rate.get()) == int(self.rate.get()):
                self.name.set(f"{self.chemical.get()} {self.rate.get():.0f} ppm")
            else:
                self.name.set(f"{self.chemical.get()} {self.rate.get():.2f} ppm")

    def clean_test(self) -> None:
        """Do some formatting on the test to clean it up for storing."""
        strippables = (self.chemical, self.name, self.label, self.clarity, self.notes)
        for attr in strippables:
            if attr.get().strip() != attr.get():
                attr.set(attr.get().strip())

    def update_label(self, *args) -> None:
        """Sets the label to the current name as a default value."""
        self.label.set(self.name.get().strip())

    def update_obs_baseline(self, *args) -> None:
        """Sets the observed baseline psi."""
        if len(self.readings) > 0:
            pressures = self.get_readings()
            self.max_psi.set(max(pressures))
            baselines = pressures[0:4]
            self.observed_baseline.set(round(sum(baselines) / 4))

    def remove_traces(self) -> None:
        """Remove tkVar traces to allow the GC to do its thing."""
        variables = (self.chemical, self.rate, self.name, self.pump_to_score)
        for var in variables:
            try:
                var.trace_remove("write", var.trace_info()[0][1])
            except IndexError:  # sometimes this spaghets on empty projects...
                pass



"""Editor for Project reporting settings."""

from __future__ import annotations

from tkinter import ttk
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from scalewiz.models.project import Project


def render(lbl: ttk.Label, ent: ttk.Entry, row: int) -> None:
    """Grids a label and entry on the passed row."""
    lbl.grid(row=row, column=0, sticky="e")
    ent.grid(row=row, column=1, sticky="ew", pady=1)


class ProjectReport(ttk.Frame):
    """Editor for Project reporting settings."""

    def __init__(self, parent: ttk.Frame, project: Project) -> None:
        super().__init__(parent)
        self.grid_columnconfigure(1, weight=1)

        lbl = ttk.Label(self, text="Export format:")
        ent = ttk.Combobox(
            self,
            values=["JSON", "CSV"],
            textvariable=project.output_format,
            state="readonly",
        )
        render(lbl, ent, 0)

        # matplotlib stuff
        # colorsLbl = ttk.Label(self, text="Plot color cycle:")
        # colorsEnt = ttk.Entry(self)
        # parent.render(colorsLbl, colorsEnt, 1)

        # legendLbl = ttk.Label(self, text="Legend location:")
        # legendEnt = ttk.Combobox(self)
        # parent.render(legendLbl, legendEnt, 2)



from __future__ import annotations

import tkinter as tk
from logging import Logger, getLogger
from tkinter import ttk
from typing import TYPE_CHECKING

from serial.tools import list_ports

if TYPE_CHECKING:
    from typing import List

    from scalewiz.models.test_handler import TestHandler

LOGGER: Logger = getLogger("scalewiz")


class DeviceBoxes(ttk.Frame):
    """A widget for selecting devices."""

    def __init__(self, parent: ttk.Frame, handler: TestHandler) -> None:
        super().__init__(parent)
        self.parent: ttk.Frame = parent
        self.handler: TestHandler = handler
        self.devices_list: List[str] = []
        self.dev1: tk.StringVar = handler.dev1
        self.dev2: tk.StringVar = handler.dev2
        self.build()

    def build(self) -> None:
        """Builds the widget."""
        # let the widgets grow to fill
        self.grid_columnconfigure(0, weight=1)
        self.grid_columnconfigure(1, weight=1)
        # make the widgets
        label = ttk.Label(self, text="     Devices:", anchor="e")
        if self.handler.is_running and not self.handler.is_done:
            state = "disabled"
        else:
            state = "normal"
        self.device1_entry = ttk.Combobox(
            self,
            width=15,
            textvariable=self.dev1,
            values=self.devices_list,
            validate="all",
            validatecommand=self.update_devices_list,
            state=state,
        )
        self.device2_entry = ttk.Combobox(
            self,
            width=15,
            textvariable=self.dev2,
            values=self.devices_list,
            validate="all",
            validatecommand=self.update_devices_list,
            state=state,
        )
        # grid the widgets
        label.grid(row=0, column=0, sticky="ne")
        self.device1_entry.grid(row=0, column=1, sticky="w")
        self.device2_entry.grid(row=0, column=2, sticky="e")
        # refresh
        self.update_devices_list()

    def update_devices_list(self, *args) -> None:
        """Updates the devices list."""
        # extra unused args are passed in by tkinter
        self.devices_list = sorted([i.device for i in list_ports.comports()])
        if len(self.devices_list) < 1:
            self.devices_list = ["None found"]

        self.device1_entry.configure(values=self.devices_list)
        self.device2_entry.configure(values=self.devices_list)

        if "None found" not in self.devices_list:
            LOGGER.debug(
                "%s found devices: %s", self.parent.handler.name, self.devices_list
            )



"""A Tkinter widget for handling tests."""

from __future__ import annotations

from logging import getLogger
from tkinter import ttk
from typing import TYPE_CHECKING

from matplotlib import pyplot as plt

from scalewiz.components.handler_view_controls import TestControls
from scalewiz.components.handler_view_devices_entry import DeviceBoxes
from scalewiz.components.handler_view_info_entry import TestInfoEntry
from scalewiz.components.handler_view_plot import LivePlot

if TYPE_CHECKING:

    from scalewiz.models.test_handler import TestHandler

LOGGER = getLogger("scalewiz")


class TestHandlerView(ttk.Frame):
    """A form for setting up / running Tests."""

    def __init__(self, parent: ttk.Frame, handler: TestHandler) -> None:
        super().__init__(parent)
        self.parent: ttk.Frame = parent
        self.handler: TestHandler = handler
        self.handler.views.append(self)
        self.plot: LivePlot = None
        self.build()

    def build(self, reload: bool = False) -> None:
        """Builds the UI, destroying any currently existing widgets."""
        if isinstance(self.plot, LivePlot):  # explicitly close to prevent memory leak
            self.after(0, plt.close, self.plot.fig)
        for child in self.winfo_children():
            self.after(0, child.destroy)
        self.grid_columnconfigure(0, weight=1)
        # row 0 ------------------------------------------------------------------------
        dev_ent = DeviceBoxes(self, self.handler)
        dev_ent.grid(row=0, column=0, sticky="new")

        # row 1 ------------------------------------------------------------------------
        frm = ttk.Frame(self)
        frm.grid_columnconfigure(1, weight=1)
        lbl = ttk.Label(frm, text="        Project:")
        lbl.grid(row=0, column=0, sticky="nw")
        proj = ttk.Label(frm, textvariable=self.handler.project.name, anchor="center")
        proj.grid(row=0, column=1, sticky="ew")
        frm.grid(row=1, column=0, sticky="new")

        # row 2 ------------------------------------------------------------------------
        test_info = TestInfoEntry(self, self.handler)
        test_info.grid(row=2, column=0, sticky="new")

        # row 3-------------------------------------------------------------------------
        test_controls = TestControls(self, self.handler)
        test_controls.grid(row=3, column=0, sticky="nsew")

        # row 0 col 1 ------------------------------------------------------------------
        plt_frm = ttk.Frame(self)
        self.plot = LivePlot(plt_frm, self.handler)
        self.plot.grid(row=0, column=0, sticky="nsew")
        plt_frm.grid(row=0, column=1, rowspan=4)

    def update_input_frame(self) -> None:
        """Disables widgets in the input frame if a Test is running."""
        if self.handler.is_running:
            for widget in self.inputs:
                widget.configure(state="disabled")
        else:
            for widget in self.inputs:
                widget.configure(state="normal")



"""Renders data from a TestHandler as it is collected."""

from __future__ import annotations

import logging
from tkinter import ttk
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.figure import SubplotParams

if TYPE_CHECKING:
    from scalewiz.models.test_handler import TestHandler

LOGGER = logging.getLogger("scalewiz")


class LivePlot(ttk.Frame):
    """Renders data from a TestHandler as it is collected."""

    def __init__(self, parent: ttk.Frame, handler: TestHandler) -> None:
        """Initialize a LivePlot."""
        super().__init__(parent)
        self.handler = handler
        self.build()

    def build(self) -> None:
        if not self.winfo_exists():
            return

        self.fig, self.axis = plt.subplots(
            figsize=(5, 3),
            dpi=100,
            constrained_layout=True,
            subplotpars=SubplotParams(left=0.1, bottom=0.1, right=0.95, top=0.95),
        )
        self.axis.set_xlabel("Time (min)")
        self.axis.set_ylabel("Pressure (psi)")
        self.axis.grid(color="darkgrey", alpha=0.65, linestyle="-")
        self.axis.set_facecolor("w")  # white
        self.fig.patch.set_facecolor("#FAFAFA")
        self.canvas = FigureCanvasTkAgg(self.fig, master=self)
        self.canvas.get_tk_widget().pack(side="top", fill="both", expand=True)
        interval = round(self.handler.project.interval_seconds.get() * 1000)  # -> ms
        self.ani = FuncAnimation(self.fig, self.animate, interval=interval)

    # could probably rewrite this with some tk.Widget.after calls
    def animate(self, interval: float) -> None:
        """Animates the live plot if a test isn't running.

        The interval argument is used by matplotlib internally.
        """
        # # we can just skip this if the test isn't running
        if len(self.handler.readings) > 0:
            if self.handler.is_running and not self.handler.is_done:
                pump1 = []
                pump2 = []
                elapsed = []  # we will share this series as an axis
                # cast to tuple in case the list changes during iteration
                readings = tuple(self.handler.readings)
                for reading in readings:
                    pump1.append(reading.pump1)
                    pump2.append(reading.pump2)
                    elapsed.append(reading.elapsedMin)
                max_psi = max((self.handler.max_psi_1, self.handler.max_psi_2))
                self.axis.clear()
                with plt.style.context("bmh"):
                    self.axis.grid(color="darkgrey", alpha=0.65, linestyle="-")
                    self.axis.set_facecolor("w")  # white
                    self.axis.set_xlabel("Time (min)")
                    self.axis.set_ylabel("Pressure (psi)")
                    self.axis.set_ylim((0, max_psi + 50))
                    self.axis.margins(0, tight=True)
                    self.axis.plot(elapsed, pump1, label="Pump 1")
                    self.axis.plot(elapsed, pump2, label="Pump 2")
                    self.axis.legend(loc="best")



"""The entry point for the program."""

import tkinter as tk

import scalewiz
from scalewiz.components.scalewiz import ScaleWiz


def main() -> None:
    """The Tkinter entry point of the program; enters mainloop."""
    root = tk.Tk()
    scalewiz.ROOT = root
    ScaleWiz(root).grid(sticky="nsew")
    root.mainloop()


if __name__ == "__main__":
    main()



"""Functions for scoring Tests within a Project.
"""
from __future__ import annotations

import tkinter as tk
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from tkinter.scrolledtext import ScrolledText
    from typing import List, Set

    from scalewiz.models.project import Project


def score(project: Project, log_widget: ScrolledText = None, *args) -> None:
    """Updates the result for every Test in the Project.

    Accepts event args passed from the tkVar trace.
    """
    # extra unused args may be passed in by tkinter
    log: List[str] = []
    log.append(f"Evaluating results for {project.name.get()}...")
    # scoring props
    limit_minutes = project.limit_minutes.get()
    interval_seconds = project.interval_seconds.get()
    max_readings = round(limit_minutes * 60 / interval_seconds)
    log.append("Max readings: limitMin * 60 / reading interval")
    log.append(f"Max readings: {max_readings}")
    baseline_area = round(project.baseline.get() * max_readings)
    log.append("Baseline area: baseline PSI * max readings")
    log.append(f"Baseline area: {project.baseline.get()} * {max_readings}")
    log.append(f"Baseline area: {baseline_area}")
    log.append("-" * 80)
    log.append("")

    # select the blanks
    blanks = []
    for test in project.tests:
        if test.is_blank.get() and test.include_on_report.get():
            blanks.append(test)
    if len(blanks) < 1:  # this is bad enough to stop us, could check earlier ..?
        return

    areas_over_blanks: Set[int] = set()
    for blank in blanks:
        log.append(f"Evaluating {blank.name.get()}")
        log.append(f"Considering data: {blank.pump_to_score.get()}")
        readings = blank.get_readings()
        log.append(f"Total readings: {len(readings)}")
        log.append(f"Observed baseline: {blank.observed_baseline.get()} psi")
        int_psi = sum(readings)
        log.append("Integral PSI: sum of all pressure readings")
        log.append(f"Integral PSI: {int_psi}")
        area = project.limit_psi.get() * len(readings) - int_psi
        log.append("Area over blank: limit_psi * # of readings - integral PSI")
        log.append(
            f"Area over blank: {project.limit_psi.get()} "
            f"* {len(readings)} - {int_psi}"
        )
        log.append(f"Area over blank: {area}")
        log.append("")
        areas_over_blanks.add(area)

    # get protectable area
    avg_blank_area = round(sum(areas_over_blanks) / len(areas_over_blanks))
    log.append(f"Avg. area over blanks: {avg_blank_area}")
    avg_protectable_area = project.limit_psi.get() * max_readings - avg_blank_area
    log.append(
        "Avg. protectable area: limit_psi * max_readings - avg. area over blanks"
    )
    log.append(
        f"Avg. protectable area: {project.limit_psi.get()} "
        f"* {max_readings} - {avg_blank_area}"
    )
    log.append(f"Avg. protectable area: {avg_protectable_area}")
    log.append("-" * 80)
    log.append("")

    # select trials
    trials = []
    for test in project.tests:
        if not test.is_blank.get():
            trials.append(test)

    # get readings
    for trial in trials:
        log.append(f"Evaluating {trial.name.get()}")
        log.append(f"Considering data: {trial.pump_to_score.get()}")
        readings = trial.get_readings()
        log.append(f"Total readings: {len(readings)}")
        log.append(f"Observed baseline: {trial.observed_baseline.get()} psi")
        int_psi = sum(readings) + (
            (max_readings - len(readings)) * project.limit_psi.get()
        )
        log.append("Integral PSI: sum of all pressure readings")
        log.append(f"Integral PSI: {int_psi}")
        result = round(1 - (int_psi - baseline_area) / avg_protectable_area, 3)
        log.append("Result: 1 - (integral PSI - baseline area) / avg protectable area")
        log.append(
            f"Result: 1 - ({int_psi} - {baseline_area}) / {avg_protectable_area}"
        )
        log.append(f"Result: {result} \n")
        trial.result.set(f"{result:.2f}")

    if isinstance(log_widget, tk.Text):
        to_log(log, log_widget)


def to_log(log: list[str], log_widget: ScrolledText) -> None:
    """Adds the passed log messages to the passed Text widget."""
    if log_widget.winfo_exists():
        log_widget.configure(state="normal")
        log_widget.delete(1.0, "end")
        for msg in log:
            log_widget.insert("end", "".join((msg, "\n")))
        log_widget.configure(state="disabled")



from __future__ import annotations

import tkinter as tk
from logging import Logger, getLogger
from tkinter import ttk
from typing import TYPE_CHECKING

from scalewiz.helpers.validation import can_be_pos_float

if TYPE_CHECKING:

    from scalewiz.models.test_handler import TestHandler


LOGGER: Logger = getLogger("scalewiz")


class TestInfoEntry(ttk.Frame):
    """A widget for inputting Test information."""

    def __init__(self, parent: tk.Widget, handler: TestHandler) -> None:
        super().__init__(parent)
        self.handler: TestHandler = handler
        self.build()

    def build(self) -> None:
        """Builds the widget."""
        self.grid_columnconfigure(1, weight=1)

        if self.handler.is_done or self.handler.is_running:
            state = "disabled"
        else:
            state = "normal"

        radio_lbl = ttk.Label(self, text="   Test Type:", anchor="e")
        radio_lbl.grid(row=0, column=0, sticky="ew")

        radio_frm = ttk.Frame(self)
        radio_frm.grid_columnconfigure(0, weight=1)
        radio_frm.grid_columnconfigure(1, weight=1)
        blank_btn = ttk.Radiobutton(
            radio_frm,
            text="Blank",
            variable=self.handler.test.is_blank,
            value=True,
            command=self.build,
            state=state,
        )
        blank_btn.grid(row=0, column=0, sticky="e", padx=25)
        trial_btn = ttk.Radiobutton(
            radio_frm,
            text="Trial",
            variable=self.handler.test.is_blank,
            value=False,
            command=self.build,
            state=state,
        )
        trial_btn.grid(row=0, column=1, sticky="w", padx=25)
        radio_frm.grid(row=0, column=1, sticky="ew")

        test_frm = ttk.Frame(self)
        test_frm.grid_columnconfigure(1, weight=1)

        if self.handler.test.is_blank.get():
            # test_frm row 0 -----------------------------------------------------------
            name_lbl = ttk.Label(test_frm, text="         Name:", anchor="e")
            name_lbl.grid(row=0, column=0, sticky="ew")
            name_ent = ttk.Entry(
                test_frm, textvariable=self.handler.test.name, state=state
            )
            name_ent.grid(row=0, column=1, sticky="ew")
            # test_frm row 1 -----------------------------------------------------------
            notes_lbl = ttk.Label(test_frm, text="Notes:", anchor="e")
            notes_lbl.grid(row=1, column=0, sticky="ew")
            if self.handler.is_running or not self.handler.is_done:
                state = "normal"
            elif self.handler.is_done:
                state = "disabled"
            notes_ent = ttk.Entry(
                test_frm, textvariable=self.handler.test.notes, state=state
            )
            notes_ent.grid(row=1, column=1, sticky="ew")
            # spacers
            ttk.Label(test_frm, text="").grid(row=2)
            ttk.Label(test_frm, text="").grid(row=3, pady=1)

        else:
            # test_frm row 0 -----------------------------------------------------------
            chem_lbl = ttk.Label(test_frm, text="Chemical:", anchor="e")
            chem_lbl.grid(row=0, column=0, sticky="e")
            chem_ent = ttk.Entry(
                test_frm, textvariable=self.handler.test.chemical, state=state
            )
            chem_ent.grid(row=0, column=1, sticky="ew")
            # test_frm row 1 -----------------------------------------------------------
            rate_lbl = ttk.Label(test_frm, text="Rate (ppm):", anchor="e")
            rate_lbl.grid(row=1, column=0, sticky="e")
            # validation command to ensure numeric inputs
            vcmd = self.register(lambda s: can_be_pos_float(s))
            rate_ent = ttk.Spinbox(
                test_frm,
                textvariable=self.handler.test.rate,
                from_=1,
                to=999999,
                validate="key",
                validatecommand=(vcmd, "%P"),
                state=state,
            )
            rate_ent.grid(row=1, column=1, sticky="ew")
            # test_frm row 2 -----------------------------------------------------------
            clarity_lbl = ttk.Label(test_frm, text="Clarity:", anchor="e")
            clarity_lbl.grid(row=2, column=0, sticky="e")
            clarity_ent = ttk.Combobox(
                test_frm,
                values=["Clear", "Slightly hazy", "Hazy"],
                textvariable=self.handler.test.clarity,
                state=state,
            )
            clarity_ent.current(0)  # default to 'Clear'
            clarity_ent.grid(row=2, column=1, sticky="ew")
            # test_frm row 3 -----------------------------------------------------------
            notes_lbl = ttk.Label(test_frm, text="Notes:", anchor="e")
            notes_lbl.grid(row=3, column=0, sticky="e")
            if self.handler.is_done:
                state = "disable"
            else:
                state = "normal"
            notes_ent = ttk.Entry(
                test_frm, textvariable=self.handler.test.notes, state=state
            )
            notes_ent.grid(row=3, column=1, sticky="ew")

        test_frm.grid(row=1, column=0, columnspan=2, sticky="ew")



"""Handles a test."""

from __future__ import annotations

import tkinter as tk
from concurrent.futures import ThreadPoolExecutor
from datetime import date
from logging import DEBUG, FileHandler, Formatter, getLogger
from pathlib import Path
from queue import Queue
from time import sleep, time
from tkinter import filedialog, messagebox
from typing import TYPE_CHECKING

from py_hplc import NextGenPump

import scalewiz
from scalewiz.models.project import Project
from scalewiz.models.test import Reading, Test

if TYPE_CHECKING:
    from logging import Logger
    from typing import List, Set, Union


class TestHandler:
    """Handles a Test."""

    # pylint: disable=too-many-instance-attributes

    def __init__(self, name: str = "Nemo") -> None:
        self.name = name
        self.root: tk.Tk = scalewiz.ROOT
        self.logger: Logger = getLogger(f"scalewiz.{name}")
        self.project: Project = Project()
        self.test: Test = None
        self.readings: List[Reading] = []
        self.max_readings: int = None  # max # of readings to collect
        self.limit_psi: int = None
        self.max_psi_1: int = None
        self.max_psi_2: int = None
        self.limit_minutes: float = None
        self.log_handler: FileHandler = None  # handles logging to log window
        self.log_queue: Queue[str] = Queue()  # view pulls from this queue
        self.dev1 = tk.StringVar()
        self.dev2 = tk.StringVar()
        self.stop_requested: bool = bool()
        self.progress = tk.IntVar()
        self.elapsed_min: float = float()  # current duration
        self.pump1: NextGenPump = None
        self.pump2: NextGenPump = None
        self.pool = ThreadPoolExecutor(max_workers=3)

        # UI concerns
        self.views: List[tk.Widget] = []  # list of views displaying the project
        self.is_running: bool = bool()
        self.is_done: bool = bool()
        self.new_test()

    @property
    def can_run(self) -> bool:
        """Returns a bool indicating whether or not the test can run."""
        return (
            (self.max_psi_1 < self.limit_psi or self.max_psi_2 < self.limit_psi)
            and self.elapsed_min < self.limit_minutes
            and len(self.readings) < self.max_readings
            and not self.stop_requested
        )

    def new_test(self) -> None:
        """Initialize a new test."""
        self.logger.info("Initializing a new test")
        if isinstance(self.test, Test):
            self.test.remove_traces()
        self.test = Test()
        self.limit_psi = self.project.limit_psi.get()
        self.limit_minutes = self.project.limit_minutes.get()
        self.max_psi_1, self.max_psi_2 = 0, 0
        self.is_running, self.is_done = False, False
        self.progress.set(0)
        self.max_readings = round(
            self.project.limit_minutes.get() * 60 / self.project.interval_seconds.get()
        )
        self.rebuild_views()

    def start_test(self) -> None:
        """Perform a series of checks to make sure the test can run, then start it."""
        issues = []
        if not Path(self.project.path.get()).is_file():
            msg = "Select an existing project file first"
            issues.append(msg)

        if self.test.name.get() == "":
            msg = "Name the experiment before starting"
            issues.append(msg)

        if self.test.name.get() in {test.name.get() for test in self.project.tests}:
            msg = "A test with this name already exists in the project"
            issues.append(msg)

        if self.test.clarity.get() == "" and not self.test.is_blank.get():
            msg = "Water clarity cannot be blank"
            issues.append(msg)

        # these methods will append issue messages if any occur
        self.update_log_handler(issues)
        self.setup_pumps(issues)
        if len(issues) > 0:
            messagebox.showwarning("Couldn't start the test", "\n".join(issues))
            for pump in (self.pump1, self.pump2):
                pump.close()
        else:
            self.readings.clear()
            self.stop_requested = False
            self.is_done = False
            self.is_running = True
            self.rebuild_views()
            self.pool.submit(self.uptake_cycle)

    def uptake_cycle(self) -> None:
        """Get ready to take readings.

        Meant to be run from a worker thread.
        """
        self.logger.info("Starting an uptake cycle")
        uptake = self.project.uptake_seconds.get()
        step = uptake / 100  # we will sleep for 100 steps
        self.pump1.run()
        self.pump2.run()
        rinse_start = time()
        for i in range(100):
            if self.can_run:
                self.progress.set(i)
                sleep(step - ((time() - rinse_start) % step))
            else:
                self.stop_test(save=False)
                break
        self.take_readings()  # still in the Future's thread

    def take_readings(self) -> None:
        """Collects Readings by messaging the pumps.

        Meant to be run from a worker thread.
        """
        self.logger.info("Starting readings collection")

        def get_pressure(pump: NextGenPump) -> Union[float, int]:
            self.logger.info("collecting a reading from %s", pump.serial.name)
            return pump.pressure

        interval = self.project.interval_seconds.get()
        start_time = time()
        # readings loop ----------------------------------------------------------------
        while self.can_run:
            self.elapsed_min = (time() - start_time) / 60
            t0 = time()
            psi1 = self.pool.submit(get_pressure, self.pump1)
            psi2 = self.pool.submit(get_pressure, self.pump2)
            psi1, psi2 = psi1.result(), psi2.result()
            t1 = time()
            self.logger.warn("got both in %s s", t1 - t0)
            average = round(((psi1 + psi2) / 2))
            reading = Reading(
                elapsedMin=self.elapsed_min, pump1=psi1, pump2=psi2, average=average
            )
            # make a message for the log in the test handler view
            msg = "@ {:.2f} min; pump1: {}, pump2: {}, avg: {}".format(
                self.elapsed_min, psi1, psi2, average
            )
            self.readings.append(reading)
            self.log_queue.put(msg)
            self.logger.debug(msg)
            prog = round((len(self.readings) / self.max_readings) * 100)
            self.progress.set(prog)

            if psi1 > self.max_psi_1:
                self.max_psi_1 = psi1
            if psi2 > self.max_psi_2:
                self.max_psi_2 = psi2

            # TYSM https://stackoverflow.com/a/25251804
            sleep(interval - ((time() - start_time) % interval))
        else:
            self.root.after(0, self.stop_test, {"save": True})

    def request_stop(self) -> None:
        """Requests that the Test stop."""
        if self.is_running:
            self.stop_requested = True

    def stop_test(self, save: bool = False, rinsing: bool = False) -> None:
        """Stops the pumps, closes their ports."""
        self.close_pumps()
        if not rinsing:
            self.is_done = True
            self.is_running = False
            for _ in range(3):
                self.views[0].bell()
        if save:
            self.save_test()
        self.progress.set(100)
        self.rebuild_views()

    def save_test(self) -> None:
        """Saves the test to the Project file in JSON format."""
        self.logger.info(
            "Saving %s to %s", self.test.name.get(), self.project.name.get()
        )
        self.test.readings.extend(self.readings)
        self.project.tests.append(self.test)
        self.project.dump_json()
        # refresh data / UI
        self.load_project(path=self.project.path.get(), new_test=False)

    def setup_pumps(self, issues: List[str] = None) -> None:
        """Set up the pumps with some default values.
        Appends errors to the passed list
        """
        if issues is None:
            issues = []

        if self.dev1.get() in ("", "None found"):
            issues.append("Select a port for pump 1")

        if self.dev2.get() in ("", "None found"):
            issues.append("Select a port for pump 2")

        if self.dev1.get() == self.dev2.get():
            issues.append("Select two unique ports")
        else:
            self.pump1 = NextGenPump(self.dev1.get(), self.logger)
            self.pump2 = NextGenPump(self.dev2.get(), self.logger)

        flowrate = self.project.flowrate.get()
        for pump in (self.pump1, self.pump2):
            if pump is None or not pump.is_open:
                issues.append(f"Couldn't connect to {pump.serial.name}")
                continue
            pump.flowrate = flowrate
        self.logger.info("Set flowrates to %s", pump.flowrate)

    def close_pumps(self) -> None:
        """Tears down the pumps."""
        for pump in (self.pump1, self.pump2):
            if pump.is_open:
                pump.stop()
                pump.close()
                self.logger.info(
                    "Stopped and closed the device @ %s",
                    pump.serial.name,
                )

    def load_project(
        self,
        path: Union[str, Path] = None,
        loaded: Set[Path] = [],
        new_test: bool = True,
    ) -> None:
        """Opens a file dialog then loads the selected Project file.

        `loaded` gets built from scratch every time it is passed in -- no need to update
        """
        if path is None:
            path = filedialog.askopenfilename(
                initialdir='C:"',
                title="Select project file:",
                filetypes=[("JSON files", "*.json")],
            )
        if isinstance(path, str):
            path = Path(path).resolve()

        # check that the dialog succeeded, the file exists, and isn't already loaded
        if path.is_file():
            if path in loaded:
                msg = "Attempted to load an already-loaded project"
                self.logger.warning(msg)
                messagebox.showwarning("Project already loaded", msg)
            else:
                self.project.remove_traces()
                self.project = Project()
                self.project.load_json(path)
                if new_test:
                    self.new_test()
                self.logger.info("Loaded %s", self.project.name.get())
                self.rebuild_views()

    def rebuild_views(self) -> None:
        """Rebuild all open Widgets that display or modify the Project file."""
        for widget in self.views:
            if widget.winfo_exists():
                self.logger.debug("Rebuilding %s", widget)
                self.root.after_idle(widget.build, {"reload": True})
            else:
                self.logger.debug("Removing dead widget %s", widget)
                self.views.remove(widget)

        self.logger.debug("Rebuilt all view widgets")

    def update_log_handler(self, issues: List[str]) -> None:
        """Sets up the logging FileHandler to the passed path."""
        id = "".join(char for char in self.test.name.get() if char.isalnum())
        log_file = f"{time():.0f}_{id}_{date.today()}.txt"
        parent_dir = Path(self.project.path.get()).parent.resolve()
        logs_dir = parent_dir.joinpath("logs").resolve()
        if not logs_dir.is_dir():
            logs_dir.mkdir()
        log_path = Path(logs_dir).joinpath(log_file).resolve()
        self.log_handler = FileHandler(log_path)

        formatter = Formatter(
            "%(asctime)s - %(thread)d - %(levelname)s - %(message)s",
            "%Y-%m-%d %H:%M:%S",
        )
        if self.log_handler in self.logger.handlers:  # remove the old one
            self.logger.removeHandler(self.log_handler)
        self.log_handler.setFormatter(formatter)
        self.log_handler.setLevel(DEBUG)
        self.logger.addHandler(self.log_handler)
        self.logger.info("Set up a log file at %s", log_file)
        self.logger.info("Starting a test for %s", self.project.name.get())



"""Displays the documentation."""

import logging
import webbrowser

LOGGER = logging.getLogger("scalewiz")


def show_help() -> None:
    """Displays the documentation in a web browser."""
    LOGGER.info("Opening the docs")
    webbrowser.open_new("https://github.com/teauxfu/scalewiz/blob/main/doc/index.rst")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['scalewiz', 'scalewiz.components', 'scalewiz.helpers', 'scalewiz.models']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0',
 'matplotlib>=3.4.2,<4.0.0',
 'pandas>=1.2.2,<2.0.0',
 'py-hplc>=1.0.1,<2.0.0',
 'tkcalendar>=1.6.1,<2.0.0',
 'tomlkit>=0.7.0,<0.8.0']

entry_points = \
{'console_scripts': ['scalewiz = scalewiz.__main__:main']}

setup_kwargs = {
    'name': 'scalewiz',
    'version': '0.5.7',
    'description': 'A graphical user interface for chemical performance testing designed to work with Teledyne SSI MX-class HPLC pumps.',
    'long_description': '========================================================================\nscalewiz |license| |python| |pypi| |build-status| |style| |code quality|\n========================================================================\n\nA graphical user interface designed to work with `Teledyne SSI MX-class\nHPLC pumps`_ for the purpose of calcite scale inhibitor chemical\nperformance testing.\n\nIf you are working with Teledyne SSI Next Generation pumps generally, please check out `py-hplc`_!\n\nThis project is stable and usable in a production environment, but listed as in beta due to the lack of a test suite (yet!).\nIf you notice something weird, fragile, or otherwise encounter a bug, please open an `issue`_.\n\n.. image:: https://raw.githubusercontent.com/teauxfu/scalewiz/main/img/main_menu.PNG\n\n.. image:: https://raw.githubusercontent.com/teauxfu/scalewiz/main/img/evaluation(plot).PNG\n\nInstallation\n============\n\n::\n\n    python -m pip install --user scalewiz\n\nOr, if you use :code:`pipx` (`try it!`_ 😉) ::\n\n    pipx install scalewiz\n\nUsage\n=====\n\n::\n\n    python -m scalewiz\n\nIf Python is on your PATH (or you used :code:`pipx` 😎), simply ::\n\n    scalewiz\n\n\nFurther instructions can be viewed in the `docs`_ section of this repo or with the Help button in the main\nmenu.\n\nAcknowledgements\n================\n- `Premier Chemical Technologies, LLC`_ for sponsoring development\n-  `@balacla`_ for support and invaluable help in brainstorming\n\n.. |license| image:: https://img.shields.io/github/license/teauxfu/scalewiz\n  :target: https://github.com/teauxfu/py-hplc/blob/main/COPYING\n  :alt: GitHub\n\n.. |python| image:: https://img.shields.io/pypi/pyversions/scalewiz\n  :alt: PyPI - Python Version\n\n.. |pypi| image:: https://img.shields.io/pypi/v/scalewiz\n  :target: https://pypi.org/project/scalewiz/\n  :alt: PyPI\n\n.. |build-status| image:: https://github.com/teauxfu/scalewiz/actions/workflows/build.yml/badge.svg\n  :target: https://github.com/teauxfu/scalewiz/actions/workflows/build.yml\n  :alt: Build Status\n\n.. |docs| image:: https://readthedocs.org/projects/pip/badge/?version=stable\n  :target: https://scalewiz.readthedocs.io/en/latest/\n  :alt: Documentation Status\n\n.. |style| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n  :target: https://github.com/psf/black\n  :alt: Style\n\n.. |code quality| image:: https://img.shields.io/badge/code%20quality-flake8-black\n  :target: https://gitlab.com/pycqa/flake8\n  :alt: Code quality\n\n.. _`Premier Chemical Technologies, LLC`: https://premierchemical.tech\n.. _`@balacla`: https://github.com/balacla\n.. _`Teledyne SSI MX-class HPLC pumps`: https://store.teledynessi.com/collections/mx-class\n.. _`py-hplc`: https://github.com/teauxfu/py-hplc\n.. _`docs`: https://github.com/teauxfu/scalewiz/blob/main/doc/index.rst#scalewiz-user-guide\n.. _`issue`: https://github.com/teauxfu/scalewiz/issues\n.. _`try it!`: https://pypa.github.io/pipx/\n',
    'author': 'Alex Whittington',
    'author_email': 'alex@southsun.tech',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""This module is for validation commands to be used with entry widgets."""


def can_be_float(s: str) -> bool:
    try:
        if float(s):
            return True
    except ValueError:
        return False


def can_be_pos_float(s: str) -> bool:
    try:
        return float(s) > 0
    except ValueError:
        return False


def can_be_pos_int(s: str) -> bool:
    try:
        return int(s) > 0
    except ValueError:
        return False



"""Returns a path to the named resource file, if it exists."""

# this is only really useful for trying to package the app in an exe

import sys
from os import path


def get_resource(name: str) -> str:
    """Returns a path to the passed file name."""
    # only really useful if trying to bundle
    # https://pyinstaller.readthedocs.io/en/stable/runtime-information.html
    bundle_dir = getattr(sys, "_MEIPASS", path.abspath(path.dirname(__file__)))
    path_to_dat = path.abspath(path.join(bundle_dir, name))
    if path.isfile(path_to_dat):
        return path_to_dat

    raise FileNotFoundError(f"Could not find a file at {path_to_dat}")



"""Simple frame that starts and stops the pumps on a timer."""

import logging
import tkinter as tk
from concurrent.futures import ThreadPoolExecutor
from time import sleep
from tkinter import ttk

from scalewiz.helpers.set_icon import set_icon
from scalewiz.models.test_handler import TestHandler

LOGGER = logging.getLogger("pct-scalewiz")


class RinseWindow(tk.Toplevel):
    """Toplevel control that starts and stops the pumps on a timer."""

    def __init__(self, handler: TestHandler) -> None:
        tk.Toplevel.__init__(self)
        self.winfo_toplevel().protocol("WM_DELETE_WINDOW", self.close)
        self.handler: TestHandler = handler
        self.pool = ThreadPoolExecutor(max_workers=1)
        self.stop = False

        set_icon(self)
        self.winfo_toplevel().title(self.handler.name)

        self.rinse_minutes = tk.IntVar()
        self.rinse_minutes.set(5)
        self.txt = tk.StringVar()
        self.txt.set("Rinse")

        # build
        lbl = ttk.Label(self, text="Rinse duration (min).:")
        lbl.grid(row=0, column=0)
        ent = ttk.Spinbox(self, textvariable=self.rinse_minutes, from_=3, to=60)
        ent.grid(row=0, column=1)

        self.button = ttk.Button(self, text="Rinse", command=self.request_rinse)
        self.button.grid(row=2, column=0, columnspan=2)

    def request_rinse(self) -> None:
        """Try to start a rinse cycle if a test isn't running."""
        if self.handler.is_done or not self.handler.is_running:
            self.pool.submit(self.rinse)

    def rinse(self) -> None:
        """Run the pumps and disable the button for the duration of a timer."""
        self.handler.setup_pumps()
        self.handler.pump1.run()
        self.handler.pump2.run()

        self.button.configure(state="disabled")
        duration = round(self.rinse_minutes.get() * 60)
        for i in range(duration):
            if not self.stop:
                self.button.configure(text=f"{i+1}/{duration} s")
                sleep(1)
            else:
                break
        self.bell()
        self.end_rinse()
        self.button.configure(state="normal", text="Rinse")

    def end_rinse(self) -> None:
        """Stop the pumps if they are running, then close their ports."""
        if self.handler.pump1.is_open:
            self.handler.pump1.stop()
            self.handler.pump1.close()
            LOGGER.info(
                "%s: Stopped and closed the device @ %s",
                self.handler.name,
                self.handler.pump1.serial.name,
            )

        if self.handler.pump2.is_open:
            self.handler.pump2.stop()
            self.handler.pump2.close()
            LOGGER.info(
                "%s: Stopped and closed the device @ %s",
                self.handler.name,
                self.handler.pump2.serial.name,
            )

    def close(self) -> None:
        """Stops the rinse cycle and closes the rinse Toplevel."""
        self.stop = True
        self.destroy()



"""Main frame widget for the application. Manages a Notebook of TestHandlerViews."""

import logging
from pathlib import Path
from tkinter import ttk

from scalewiz.components.handler_view import TestHandlerView
from scalewiz.components.scalewiz_menu_bar import MenuBar
from scalewiz.models.test_handler import TestHandler

LOGGER = logging.getLogger("scalewiz")


class MainFrame(ttk.Frame):
    """Main Frame for the application."""

    def __init__(self, parent: ttk.Frame) -> None:
        super().__init__(parent)
        self.winfo_toplevel().protocol("WM_DELETE_WINDOW", self.close)
        self.build()

    def build(self) -> None:
        """Build the UI."""
        self.winfo_toplevel().configure(menu=MenuBar(self).menubar)
        self.tab_control: ttk.Notebook = ttk.Notebook(self)
        self.tab_control.grid(sticky="nsew")
        self.add_handler()

    def add_handler(self) -> None:
        """Adds a new tab with an associated test handler."""
        system_name = f"  System {len(self.tab_control.tabs())+1}  "
        handler = TestHandler(name=system_name.strip())
        self.tab_control.add(
            TestHandlerView(self.tab_control, handler), sticky="nsew", text=system_name
        )
        LOGGER.info("Added %s to main window", handler.name)
        # if this is the first handler, open the most recent project
        if len(self.tab_control.tabs()) == 1:
            from scalewiz import CONFIG

            recent = CONFIG["recents"]["project"]
            if recent != "":
                recent = Path(recent)
                if recent.is_file():
                    handler.load_project(recent)

    def close(self) -> None:
        """Closes the program if no tests are running."""
        for tab in self.tab_control.tabs():
            widget: TestHandlerView = self.nametowidget(tab)
            if widget.handler.is_running:
                LOGGER.warning(
                    "Attempted to close while a test was running on %s",
                    widget.handler.name,
                )
                return
        self.quit()



"""The parent module for the scalewiz package."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from tkinter import Tk

from scalewiz.helpers.configuration import get_config

ROOT: Tk = None
CONFIG: dict = get_config()
